from __future__ import absolute_import
from typing import Any, Union, Optional, Dict, Mapping


class Status:
    UP = 'UP'
    DOWN = 'DOWN'
    SARTING = 'STARTING'
    OUT_OF_SERVICE = 'OUT_OF_SERVICE'
    UNKNOWN = 'UNKNOWN'

    code: str
    description: Optional[str]

    def __init__(self, code: str, description: Optional[str] = None):
        self.code = code
        self.description = description


HealthStatus = Union[str, Status]


class Health:
    _status: Status
    _details: Dict[str, Any]

    def __init__(self, status: Status, details: Mapping[str, Any]):
        self._status = status
        self._details = dict(details) if details else {}

    def get_status(self) -> Status:
        return self._status

    def get_details(self) -> Mapping[str, Any]:
        return self._details

    def __json__(self):
        json = {
            'status': self._status.code,
            'description': self._status.description
        }

        for key, value in self._details.items():
            if hasattr(value, '__json__'):
                json[key] = value.__json__()
            else:
                json[key] = value
        return json

    @staticmethod
    def unknown(description: Optional[str] = None):
        s = Status(Status.UNKNOWN, description)
        return HealthBuilder(s)

    @staticmethod
    def down(description: Optional[str] = None):
        s = Status(Status.DOWN, description)
        return HealthBuilder(s)

    @staticmethod
    def up(description: Optional[str] = None):
        s = Status(Status.UP, description)
        return HealthBuilder(s)

    @staticmethod
    def out_of_service(description: Optional[str] = None):
        s = Status(Status.OUT_OF_SERVICE, description)
        return HealthBuilder(s)

    @staticmethod
    def status(code: str, description: Optional[str] = None):
        s = Status(code, description)
        return HealthBuilder(s)

    @staticmethod
    def new_builder(status: HealthStatus, details: Optional[Mapping[str, Any]] = None) -> "HealthBuilder":
        return HealthBuilder(status, details)


class HealthBuilder:
    def __init__(self, status: HealthStatus, details: Optional[Mapping[str, Any]] = None):
        self.status(status)
        self._details = dict(details) if details else {}

    def status(self, status: HealthStatus, **kwargs):
        if status:
            if isinstance(status, Status):
                self._status = status
            else:
                description: Optional[str] = kwargs.get('description')
                self._status = Status(status, description)
        return self

    def down(self, message: str):
        self._status = Status(Status.DOWN, message)
        return self

    def add_detail(self, name: str, item: Any):
        self._details[name] = item
        return self

    def build(self):
        return Health(self._status, self._details)



from __future__ import absolute_import
from typing import Dict
from mctech_core import Configure
from ...health import get_health_manager, MetricEndPoint

METRICS_PATH = '/metrics'


def create_metrics_routers(configure: Configure) -> Dict[str, MetricEndPoint]:
    routers: Dict[str, MetricEndPoint] = {}
    for m in get_health_manager().get_metrics():
        routers[METRICS_PATH + m.path] = m.endpoint
    return routers



from __future__ import absolute_import
from .health import Health, HealthBuilder, Status, CompositeHealthIndicator, Indicator, MetricIndicator, MetricEndPoint
from .actuator import create_actuator_route

from .health import get_health_manager


class DefaultIndicator(Indicator):
    @property
    def name(self) -> str:
        return 'application'

    def health(self) -> Health:
        return Health.up().build()


get_health_manager().add_indicator(DefaultIndicator())

__all__ = [
    "Health",
    "HealthBuilder",
    "Status",
    "get_health_manager",
    "CompositeHealthIndicator",
    "Indicator",
    "MetricIndicator",
    "MetricEndPoint",
    "create_actuator_route"
]



from __future__ import absolute_import
import abc

from typing import Callable, Coroutine, NamedTuple, Any
from fastapi import Response
from .health import Health

MetricEndPoint = Callable[..., Coroutine[Any, Any, Response]]


class Indicator:
    @property
    @abc.abstractmethod
    def name(self) -> str:
        pass

    @abc.abstractmethod
    def health(self) -> Health:
        pass


class MetricIndicator(NamedTuple):
    path: str
    endpoint: MetricEndPoint



from __future__ import absolute_import
from mctech_core import Configure
from fastapi.responses import JSONResponse


def create_info_route(configure: Configure):
    async def info():
        return JSONResponse({
            'app': configure.get_app_info()
        })

    return info



from __future__ import absolute_import
from .info_route import create_info_route
from .env_route import create_env_route
from .health_route import create_health_route
from .metrics_routes import create_metrics_routers

__all__ = [
    "create_info_route",
    "create_env_route",
    "create_health_route",
    "create_metrics_routers"
]



from setuptools import setup, find_packages

setup(
    name="mctech_actuator",
    version="1.1.1",
    packages=find_packages(
        include=["mctech_actuator**"],
        exclude=["*.tests"]
    ),
    exclude_package_data={"mctech_actuator": ["testmain.py"]},
    install_requires=["log4py", "fastapi"]
)



from __future__ import absolute_import
from typing import List
from .composite_health_indicator import CompositeHealthIndicator
from .indicator import Indicator, MetricIndicator
from .health import Health


class HealthManager:
    def __init__(self):
        self._indicator = CompositeHealthIndicator()
        self._metrics: List[MetricIndicator] = []

    def get_metrics(self) -> List[MetricIndicator]:
        return self._metrics

    def add_metric(self, metric: MetricIndicator):
        self._metrics.append(metric)

    def get_health(self) -> Health:
        return self._indicator.health()

    def add_indicator(self, indicator: Indicator):
        holder = IndicatorHolder(indicator)
        self._indicator.add_indicator(indicator.name, holder)


class IndicatorHolder:
    def __init__(self, indicator):
        self.name = None
        self.indicator = indicator

    def health(self):
        try:
            h = self.indicator.health()
        except BaseException as e:
            h = Health.down(type(e).__name__ + ':' + str(e)).build()
        return h


_manager = HealthManager()


def get_health_manager() -> HealthManager:
    return _manager




from __future__ import absolute_import
from fastapi.responses import JSONResponse
from mctech_core import Configure

from ...health import get_health_manager


def create_health_route(configure: Configure):
    async def health():
        try:
            health = get_health_manager.get_health()
            result = health.__json__()
        except Exception as e:
            result = {
                'status': 'DOWN',
                'error': str(e)
            }
        return JSONResponse(result)
    return health




from __future__ import absolute_import
from fastapi import FastAPI
from fastapi.responses import PlainTextResponse
from mctech_core import Configure
from .route import create_info_route, create_env_route, create_health_route, create_metrics_routers


def create_actuator_route(configure: Configure, app: FastAPI):
    async def empty():
        res = PlainTextResponse('Actuator Not Found')
        res.status_code = 404
        return res

    management = configure.get_config('management')
    prefix = management.get('context-path') or \
        management.get('contextPath') or '/actuator'

    app.add_api_route(f"{prefix}/info", endpoint=create_info_route(configure))
    app.add_api_route(f"{prefix}/health", endpoint=create_health_route(configure))
    app.add_api_route(f"{prefix}/env", endpoint=create_env_route(configure))
    app.add_api_route(f"{prefix}/shutdown", endpoint=empty)

    app.add_api_route(f"{prefix}/loggers", endpoint=empty)
    app.add_api_route(f"{prefix}/loggers", methods=['head'], endpoint=empty)
    app.add_api_route(f"{prefix}/loggers/:category", endpoint=empty)

    metric_routers = create_metrics_routers(configure)
    for path, endpoint in metric_routers.items():
        app.add_api_route(f"{prefix}/{path}", endpoint=endpoint)


__all__ = [
    "create_actuator_route"
]



from __future__ import absolute_import
from typing import Dict, Mapping, Any, List, Optional
from .health import Health, Status
from .indicator import Indicator


class CompositeHealthIndicator(Indicator):
    _indicators: Dict[str, Indicator]
    _health_aggregator: "HealthAggregator"

    def __init__(self, indicators: Optional[Mapping[str, Indicator]] = None):
        self._indicators = dict(indicators) if indicators else {}
        self._health_aggregator = HealthAggregator()

    def add_indicator(self, name: str, indicator):
        self._indicators[name] = indicator

    def health(self) -> Health:
        healths: Dict[str, Health] = {}
        for key, value in self._indicators.items():
            try:
                healths[key] = value.health()
            except Exception as ex:
                healths[key] = Health.unknown("ERROR: " + str(ex)).build()
        return self._health_aggregator.aggregate(healths)

    @property
    def name(self) -> str:
        return "composte"


__STATUS_ORDERS = ['DOWN', 'OUT_OF_SERVICE', 'UP', 'UNKNOWN']


def aggregate_status(candidates: List[Status]):
    # Only sort those status instances that we know about
    filtered_candidates: List[Status] = []
    for candidate in candidates:
        if next(
                filter(lambda code: candidate.code == code, __STATUS_ORDERS),
                None):
            filtered_candidates.append(candidate)

    # If no status is given return UNKNOWN
    if len(filtered_candidates) == 0:
        return Status.UNKNOWN

    if len(filtered_candidates) == 1:
        return filtered_candidates[0]

    # Sort given Status instances by configured order
    filtered_candidates.sort(key=lambda s: __STATUS_ORDERS.index(s.code))
    return filtered_candidates[0]


def aggregate_details(healths: Dict[str, Health]) -> Dict[str, Any]:
    return dict(healths)


class HealthAggregator:
    def aggregate(self, healths: Dict[str, Health]) -> Health:
        status_candidates: List[Status] = []
        for health in healths.values():
            status_candidates.append(health.get_status())
        status = aggregate_status(status_candidates)
        details = aggregate_details(healths)
        return Health.new_builder(status, details).build()



from __future__ import absolute_import
import sys
import os

from typing import Dict, List, Any
from fastapi.responses import JSONResponse

from mctech_core import Configure

SECURE_PROP_NAMES = [
    'admin',
    'user',
    'password',
    'pass',
    'pwd',
    'login',
    'username',
    'secret',
    'token',
    'aes'
]


def _prapare_object(src: Any, target: Any, prefix: str):
    if hasattr(src, '__json__'):
        src = src.__json__()

    if isinstance(src, Dict):
        for name, value in src.items():
            # 可能含有敏感信息的字符串属性
            # 递归处理
            _prapare_object(value, target, prefix + name + '.')
    elif isinstance(src, List):
        for i in range(0, len(src)):
            item_value = src[i]
            _prapare_object(item_value, target, f"{prefix}[{i}].")
    else:
        if prefix.endswith('.'):
            prefix = prefix[:-1]

        lower_prefix = prefix.lower()
        match = next(filter(lambda p: p in lower_prefix, SECURE_PROP_NAMES), None)
        if match and isinstance(src, str):
            # hide secure details
            src = '*******'
        target[prefix] = src


def create_env_route(configure: Configure):
    async def health():
        config = {}
        configureSources = {}
        environment = {}
        args = {}
        _prapare_object(configure.get_config(''), config, '')
        # prapareObject(configure.getConfigSourceInfo(), configureSources, '')
        _prapare_object(dict(os.environ), environment, '')
        _prapare_object(list(sys.argv), args, '')

        result = {
            'config': config,
            'configureSources': configureSources,
            'environment': environment,
            'args': args
        }
        return JSONResponse(result)
    return health



from __future__ import absolute_import
from .health import Health, Status, HealthBuilder
from .composite_health_indicator import CompositeHealthIndicator
from .health_manager import get_health_manager
from .indicator import Indicator, MetricIndicator, MetricEndPoint


__all__ = [
    "Health",
    "HealthBuilder",
    "Status",
    "CompositeHealthIndicator",
    "get_health_manager",
    "Indicator",
    "MetricIndicator",
    "MetricEndPoint"
]
